# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import List, NewType, Optional, TypeVar

import pytest

import sciline as sl
from sciline.typing import Item, Label


def test_set_param_table_raises_if_param_names_are_duplicate() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    with pytest.raises(ValueError):
        pl.set_param_table(sl.ParamTable(str, {float: [4.0, 5.0, 6.0]}))
    assert pl.compute(Item((Label(int, 1),), float)) == 2.0
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(Item((Label(str, 1),), float))


def test_set_param_table_raises_if_row_dim_is_duplicate() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    with pytest.raises(ValueError):
        pl.set_param_table(sl.ParamTable(int, {str: ['a', 'b', 'c']}))
    assert pl.compute(Item((Label(int, 1),), float)) == 2.0
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(Item((Label(int, 1),), str))


def test_can_get_elements_of_param_table() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(Item((Label(int, 1),), float)) == 2.0


def test_can_get_elements_of_param_table_with_explicit_index() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}, index=[11, 12, 13]))
    assert pl.compute(Item((Label(int, 12),), float)) == 2.0


def test_can_depend_on_elements_of_param_table() -> None:
    # This is not a valid type annotation, notsure why it works with get_type_hints
    def use_elem(x: Item((Label(int, 1),), float)) -> str:  # type: ignore[valid-type]
        return str(x)

    pl = sl.Pipeline([use_elem])
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(str) == "2.0"


def test_can_compute_series_of_param_values() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(sl.Series[int, float]) == sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})


def test_cannot_compute_series_of_non_table_param() -> None:
    pl = sl.Pipeline()
    # Table for defining length
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    pl[str] = 'abc'
    # The alternative option would be to expect to return
    #     sl.Series(int, {0: 'abc', 1: 'abc', 2: 'abc'})
    # For now, we are not supporting this since it is unclear if this would be
    # conceptually sound and risk free.
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(sl.Series[int, str])


def test_can_compute_series_of_derived_values() -> None:
    def process(x: float) -> str:
        return str(x)

    pl = sl.Pipeline([process])
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(sl.Series[int, str]) == sl.Series(
        int, {0: "1.0", 1: "2.0", 2: "3.0"}
    )


def test_creating_pipeline_with_provider_of_series_raises() -> None:
    series = sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})

    def make_series() -> sl.Series[int, float]:
        return series

    with pytest.raises(ValueError):
        sl.Pipeline([make_series])


def test_creating_pipeline_with_series_param_raises() -> None:
    series = sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})

    with pytest.raises(ValueError):
        sl.Pipeline([], params={sl.Series[int, float]: series})


def test_explicit_index_of_param_table_is_forwarded_correctly() -> None:
    def process(x: float) -> int:
        return int(x)

    pl = sl.Pipeline([process])
    pl.set_param_table(
        sl.ParamTable(str, {float: [1.0, 2.0, 3.0]}, index=['a', 'b', 'c'])
    )
    assert pl.compute(sl.Series[str, int]) == sl.Series(str, {'a': 1, 'b': 2, 'c': 3})


def test_can_gather_index() -> None:
    Sum = NewType("Sum", float)
    Name = NewType("Name", str)

    def gather(x: sl.Series[Name, float]) -> Sum:
        return Sum(sum(x.values()))

    def make_float(x: str) -> float:
        return float(x)

    pl = sl.Pipeline([gather, make_float])
    pl.set_param_table(sl.ParamTable(Name, {str: ["1.0", "2.0", "3.0"]}))
    assert pl.compute(Sum) == 6.0


def test_can_zip() -> None:
    Sum = NewType("Sum", str)
    Str = NewType("Str", str)
    Run = NewType("Run", int)

    def gather_zip(x: sl.Series[Run, Str], y: sl.Series[Run, int]) -> Sum:
        z = [f'{x_}{y_}' for x_, y_ in zip(x.values(), y.values())]
        return Sum(str(z))

    def use_str(x: str) -> Str:
        return Str(x)

    pl = sl.Pipeline([gather_zip, use_str])
    pl.set_param_table(sl.ParamTable(Run, {str: ['a', 'a', 'ccc'], int: [1, 2, 3]}))

    assert pl.compute(Sum) == "['a1', 'a2', 'ccc3']"


def test_diamond_dependency_pulls_values_from_columns_in_same_param_table() -> None:
    Sum = NewType("Sum", float)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Row = NewType("Row", int)

    def gather(x: sl.Series[Row, float]) -> Sum:
        return Sum(sum(x.values()))

    def join(x: Param1, y: Param2) -> float:
        return x / y

    pl = sl.Pipeline([gather, join])
    pl.set_param_table(sl.ParamTable(Row, {Param1: [1, 4, 9], Param2: [1, 2, 3]}))

    assert pl.compute(Sum) == Sum(6)


def test_diamond_dependency_on_same_column() -> None:
    Sum = NewType("Sum", float)
    Param = NewType("Param", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Row = NewType("Row", int)

    def gather(x: sl.Series[Row, float]) -> Sum:
        return Sum(sum(x.values()))

    def to_param1(x: Param) -> Param1:
        return Param1(x)

    def to_param2(x: Param) -> Param2:
        return Param2(x)

    def join(x: Param1, y: Param2) -> float:
        return x / y

    pl = sl.Pipeline([gather, join, to_param1, to_param2])
    pl.set_param_table(sl.ParamTable(Row, {Param: [1, 2, 3]}))

    assert pl.compute(Sum) == Sum(3)


def test_dependencies_on_different_param_tables_broadcast() -> None:
    Row1 = NewType("Row1", int)
    Row2 = NewType("Row2", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Product = NewType("Product", str)

    def gather_both(x: sl.Series[Row1, Param1], y: sl.Series[Row2, Param2]) -> Product:
        broadcast = [[x_, y_] for x_ in x.values() for y_ in y.values()]
        return Product(str(broadcast))

    pl = sl.Pipeline([gather_both])
    pl.set_param_table(sl.ParamTable(Row1, {Param1: [1, 2, 3]}))
    pl.set_param_table(sl.ParamTable(Row2, {Param2: [4, 5]}))
    assert pl.compute(Product) == "[[1, 4], [1, 5], [2, 4], [2, 5], [3, 4], [3, 5]]"


def test_dependency_on_other_param_table_in_parent_broadcasts_branch() -> None:
    Row1 = NewType("Row1", int)
    Row2 = NewType("Row2", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Summed2 = NewType("Summed2", int)
    Product = NewType("Product", str)

    def gather2_and_combine(x: Param1, y: sl.Series[Row2, Param2]) -> Summed2:
        return Summed2(x * sum(y.values()))

    def gather1(x: sl.Series[Row1, Summed2]) -> Product:
        return Product(str(list(x.values())))

    pl = sl.Pipeline([gather1, gather2_and_combine])
    pl.set_param_table(sl.ParamTable(Row1, {Param1: [1, 2, 3]}))
    pl.set_param_table(sl.ParamTable(Row2, {Param2: [4, 5]}))
    assert pl.compute(Product) == "[9, 18, 27]"


def test_dependency_on_other_param_table_in_grandparent_broadcasts_branch() -> None:
    Row1 = NewType("Row1", int)
    Row2 = NewType("Row2", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Summed2 = NewType("Summed2", int)
    Combined = NewType("Combined", int)
    Product = NewType("Product", str)

    def gather2(x: sl.Series[Row2, Param2]) -> Summed2:
        return Summed2(sum(x.values()))

    def combine(x: Param1, y: Summed2) -> Combined:
        return Combined(x * y)

    def gather1(x: sl.Series[Row1, Combined]) -> Product:
        return Product(str(list(x.values())))

    pl = sl.Pipeline([gather1, gather2, combine])
    pl.set_param_table(sl.ParamTable(Row1, {Param1: [1, 2, 3]}))
    pl.set_param_table(sl.ParamTable(Row2, {Param2: [4, 5]}))
    assert pl.compute(Product) == "[9, 18, 27]"


def test_nested_dependencies_on_different_param_tables() -> None:
    Row1 = NewType("Row1", int)
    Row2 = NewType("Row2", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Combined = NewType("Combined", int)

    def combine(x: Param1, y: Param2) -> Combined:
        return Combined(x * y)

    pl = sl.Pipeline([combine])
    pl.set_param_table(sl.ParamTable(Row1, {Param1: [1, 2, 3]}))
    pl.set_param_table(sl.ParamTable(Row2, {Param2: [4, 5]}))
    assert pl.compute(sl.Series[Row1, sl.Series[Row2, Combined]]) == sl.Series(
        Row1,
        {
            0: sl.Series(Row2, {0: 4, 1: 5}),
            1: sl.Series(Row2, {0: 8, 1: 10}),
            2: sl.Series(Row2, {0: 12, 1: 15}),
        },
    )
    assert pl.compute(sl.Series[Row2, sl.Series[Row1, Combined]]) == sl.Series(
        Row2,
        {
            0: sl.Series(Row1, {0: 4, 1: 8, 2: 12}),
            1: sl.Series(Row1, {0: 5, 1: 10, 2: 15}),
        },
    )


def test_can_groupby_by_requesting_series_of_series() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)

    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(Row, {Param1: [1, 1, 3], Param2: [4, 5, 6]}))
    expected = sl.Series(
        Param1,
        {1: sl.Series(Row, {0: 4, 1: 5}), 3: sl.Series(Row, {2: 6})},
    )
    assert pl.compute(sl.Series[Param1, sl.Series[Row, Param2]]) == expected


def test_groupby_by_requesting_series_of_series_preserves_indices() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)

    pl = sl.Pipeline()
    pl.set_param_table(
        sl.ParamTable(Row, {Param1: [1, 1, 3], Param2: [4, 5, 6]}, index=[11, 12, 13])
    )
    assert pl.compute(sl.Series[Param1, sl.Series[Row, Param2]]) == sl.Series(
        Param1, {1: sl.Series(Row, {11: 4, 12: 5}), 3: sl.Series(Row, {13: 6})}
    )


def test_can_groupby_by_param_used_in_ancestor() -> None:
    Row = NewType("Row", int)
    Param = NewType("Param", str)

    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(Row, {Param: ['x', 'x', 'y']}))
    expected = sl.Series(
        Param,
        {"x": sl.Series(Row, {0: "x", 1: "x"}), "y": sl.Series(Row, {2: "y"})},
    )
    assert pl.compute(sl.Series[Param, sl.Series[Row, Param]]) == expected


def test_multi_level_groupby_raises_with_params_from_same_table() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    pl.set_param_table(
        sl.ParamTable(
            Row, {Param1: [1, 1, 1, 3], Param2: [4, 5, 5, 6], Param3: [7, 8, 9, 10]}
        )
    )
    with pytest.raises(ValueError, match='Could not find unique grouping node'):
        pl.compute(sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]])


def test_multi_level_groupby_with_params_from_different_table() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    grouping1 = sl.ParamTable(Row, {Param2: [0, 1, 1, 2], Param3: [7, 8, 9, 10]})
    # We are not providing an explicit index here, so this only happens to work because
    # the values of Param2 match range(2).
    grouping2 = sl.ParamTable(Param2, {Param1: [1, 1, 3]})
    pl.set_param_table(grouping1)
    pl.set_param_table(grouping2)
    assert pl.compute(
        sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]]
    ) == sl.Series(
        Param1,
        {
            1: sl.Series(
                Param2, {0: sl.Series(Row, {0: 7}), 1: sl.Series(Row, {1: 8, 2: 9})}
            ),
            3: sl.Series(Param2, {2: sl.Series(Row, {3: 10})}),
        },
    )


def test_multi_level_groupby_with_params_from_different_table_can_select() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    grouping1 = sl.ParamTable(Row, {Param2: [4, 5, 5, 6], Param3: [7, 8, 9, 10]})
    # Note the missing index "6" here.
    grouping2 = sl.ParamTable(Param2, {Param1: [1, 1]}, index=[4, 5])
    pl.set_param_table(grouping1)
    pl.set_param_table(grouping2)
    assert pl.compute(
        sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]]
    ) == sl.Series(
        Param1,
        {
            1: sl.Series(
                Param2, {4: sl.Series(Row, {0: 7}), 5: sl.Series(Row, {1: 8, 2: 9})}
            )
        },
    )


def test_multi_level_groupby_with_params_from_different_table_preserves_index() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    grouping1 = sl.ParamTable(
        Row, {Param2: [4, 5, 5, 6], Param3: [7, 8, 9, 10]}, index=[100, 200, 300, 400]
    )
    grouping2 = sl.ParamTable(Param2, {Param1: [1, 1, 3]}, index=[4, 5, 6])
    pl.set_param_table(grouping1)
    pl.set_param_table(grouping2)
    assert pl.compute(
        sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]]
    ) == sl.Series(
        Param1,
        {
            1: sl.Series(
                Param2,
                {4: sl.Series(Row, {100: 7}), 5: sl.Series(Row, {200: 8, 300: 9})},
            ),
            3: sl.Series(Param2, {6: sl.Series(Row, {400: 10})}),
        },
    )


def test_multi_level_groupby_with_params_from_different_table_can_reorder() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    grouping1 = sl.ParamTable(
        Row, {Param2: [4, 5, 5, 6], Param3: [7, 8, 9, 10]}, index=[100, 200, 300, 400]
    )
    grouping2 = sl.ParamTable(Param2, {Param1: [1, 1, 3]}, index=[6, 5, 4])
    pl.set_param_table(grouping1)
    pl.set_param_table(grouping2)
    assert pl.compute(
        sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]]
    ) == sl.Series(
        Param1,
        {
            1: sl.Series(
                Param2,
                {6: sl.Series(Row, {400: 10}), 5: sl.Series(Row, {200: 8, 300: 9})},
            ),
            3: sl.Series(Param2, {4: sl.Series(Row, {100: 7})}),
        },
    )


@pytest.mark.parametrize("index", [None, [4, 5, 7]])
def test_multi_level_groupby_raises_on_index_mismatch(
    index: Optional[List[int]],
) -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)
    Param3 = NewType("Param3", int)

    pl = sl.Pipeline()
    grouping1 = sl.ParamTable(
        Row, {Param2: [4, 5, 5, 6], Param3: [7, 8, 9, 10]}, index=[100, 200, 300, 400]
    )
    grouping2 = sl.ParamTable(Param2, {Param1: [1, 1, 3]}, index=index)
    pl.set_param_table(grouping1)
    pl.set_param_table(grouping2)
    with pytest.raises(ValueError):
        pl.compute(sl.Series[Param1, sl.Series[Param2, sl.Series[Row, Param3]]])


@pytest.mark.parametrize("index", [None, [4, 5, 6]])
def test_groupby_over_param_table(index: Optional[List[int]]) -> None:
    Index = NewType("Index", int)
    Name = NewType("Name", str)
    Param = NewType("Param", int)
    ProcessedParam = NewType("ProcessedParam", int)
    SummedGroup = NewType("SummedGroup", int)
    ProcessedGroup = NewType("ProcessedGroup", int)

    def process_param(x: Param) -> ProcessedParam:
        return ProcessedParam(x + 1)

    def sum_group(group: sl.Series[Index, ProcessedParam]) -> SummedGroup:
        return SummedGroup(sum(group.values()))

    def process(x: SummedGroup) -> ProcessedGroup:
        return ProcessedGroup(2 * x)

    params = sl.ParamTable(
        Index, {Param: [1, 2, 3], Name: ['a', 'a', 'b']}, index=index
    )
    pl = sl.Pipeline([process_param, sum_group, process])
    pl.set_param_table(params)

    graph = pl.get(sl.Series[Name, ProcessedGroup])
    assert graph.compute() == sl.Series(Name, {'a': 10, 'b': 8})


def test_requesting_series_index_that_is_not_in_param_table_raises() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)

    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(Row, {Param1: [1, 1, 3], Param2: [4, 5, 6]}))

    with pytest.raises(KeyError):
        pl.compute(sl.Series[int, Param2])


def test_requesting_series_index_that_is_a_param_raises_if_not_grouping() -> None:
    Row = NewType("Row", int)
    Param1 = NewType("Param1", int)
    Param2 = NewType("Param2", int)

    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(Row, {Param1: [1, 1, 3], Param2: [4, 5, 6]}))
    with pytest.raises(ValueError, match='Could not find unique grouping node'):
        pl.compute(sl.Series[Param1, Param2])


def test_generic_providers_work_with_param_tables() -> None:
    Param = TypeVar('Param')
    Row = NewType("Row", int)

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized(x: Param) -> Str[Param]:
        return Str(f'{x}')

    def make_float() -> float:
        return 1.5

    pipeline = sl.Pipeline([make_float, parametrized])
    pipeline.set_param_table(sl.ParamTable(Row, {int: [1, 2, 3]}))

    assert pipeline.compute(Str[float]) == Str[float]('1.5')
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(Str[int])
    assert pipeline.compute(sl.Series[Row, Str[int]]) == sl.Series(
        Row,
        {
            0: Str[int]('1'),
            1: Str[int]('2'),
            2: Str[int]('3'),
        },
    )


def test_generic_provider_can_depend_on_param_series() -> None:
    Param = TypeVar('Param')
    Row = NewType("Row", int)

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized_gather(x: sl.Series[Row, Param]) -> Str[Param]:
        return Str(f'{list(x.values())}')

    pipeline = sl.Pipeline([parametrized_gather])
    pipeline.set_param_table(
        sl.ParamTable(Row, {int: [1, 2, 3], float: [1.5, 2.5, 3.5]})
    )

    assert pipeline.compute(Str[int]) == Str[int]('[1, 2, 3]')
    assert pipeline.compute(Str[float]) == Str[float]('[1.5, 2.5, 3.5]')


def test_generic_provider_can_depend_on_derived_param_series() -> None:
    T = TypeVar('T')
    Row = NewType("Row", int)

    class Str(sl.Scope[T, str], str):
        ...

    def use_param(x: int) -> float:
        return x + 0.5

    def parametrized_gather(x: sl.Series[Row, T]) -> Str[T]:
        return Str(f'{list(x.values())}')

    pipeline = sl.Pipeline([parametrized_gather, use_param])
    pipeline.set_param_table(sl.ParamTable(Row, {int: [1, 2, 3]}))

    assert pipeline.compute(Str[float]) == Str[float]('[1.5, 2.5, 3.5]')


def test_params_in_table_can_be_generic() -> None:
    T = TypeVar('T')
    Row = NewType("Row", int)

    class Str(sl.Scope[T, str], str):
        ...

    class Param(sl.Scope[T, str], str):
        ...

    def parametrized_gather(x: sl.Series[Row, Param[T]]) -> Str[T]:
        return Str(','.join(x.values()))

    pipeline = sl.Pipeline([parametrized_gather])
    pipeline.set_param_table(
        sl.ParamTable(Row, {Param[int]: ["1", "2"], Param[float]: ["1.5", "2.5"]})
    )

    assert pipeline.compute(Str[int]) == Str[int]('1,2')
    assert pipeline.compute(Str[float]) == Str[float]('1.5,2.5')


def test_compute_time_handler_works_alongside_param_table() -> None:
    Missing = NewType("Missing", str)

    def process(x: float, missing: Missing) -> str:
        return str(x) + missing

    pl = sl.Pipeline([process])
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    pl.get(sl.Series[int, str], handler=sl.HandleAsComputeTimeException())


def test_param_table_column_and_param_of_same_type_can_coexist() -> None:
    pl = sl.Pipeline()
    pl[float] = 1.0
    pl.set_param_table(sl.ParamTable(int, {float: [2.0, 3.0]}))
    assert pl.compute(float) == 1.0
    assert pl.compute(sl.Series[int, float]) == sl.Series(int, {0: 2.0, 1: 3.0})


def test_pipeline_copy_with_param_table() -> None:
    a = sl.Pipeline()
    a.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    b = a.copy()
    assert b.compute(sl.Series[int, float]) == sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})


def test_pipeline_set_param_table_on_original_does_not_affect_copy() -> None:
    a = sl.Pipeline()
    b = a.copy()
    a.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert a.compute(sl.Series[int, float]) == sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})
    with pytest.raises(sl.UnsatisfiedRequirement):
        b.compute(sl.Series[int, float])


def test_pipeline_set_param_table_on_copy_does_not_affect_original() -> None:
    a = sl.Pipeline()
    b = a.copy()
    b.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert b.compute(sl.Series[int, float]) == sl.Series(int, {0: 1.0, 1: 2.0, 2: 3.0})
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(sl.Series[int, float])


def test_can_make_html_repr_with_param_table() -> None:
    pl = sl.Pipeline()
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl._repr_html_()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType, Optional, Union

import pytest

import sciline as sl


def test_provider_returning_optional_disallowed() -> None:
    def make_optional() -> Optional[int]:
        return 3

    with pytest.raises(ValueError):
        sl.Pipeline([make_optional])


def test_provider_returning_union_disallowed() -> None:
    def make_union() -> Union[int, float]:
        return 3

    with pytest.raises(ValueError):
        sl.Pipeline([make_union])


def test_parameter_type_union_or_optional_disallowed() -> None:
    pipeline = sl.Pipeline()
    with pytest.raises(ValueError):
        pipeline[Union[int, float]] = 3  # type: ignore[index]
    with pytest.raises(ValueError):
        pipeline[Optional[int]] = 3  # type: ignore[index]


def test_union_requirement_satisfied_by_unique_match() -> None:
    def require_union(x: Union[int, float]) -> str:
        return f'{x}'

    pipeline = sl.Pipeline([require_union])
    pipeline[int] = 1
    assert pipeline.compute(str) == '1'


def test_union_requirement_with_multiple_matches_raises_AmbiguousProvider() -> None:
    def require_union(x: Union[int, float]) -> str:
        return f'{x}'

    pipeline = sl.Pipeline([require_union])
    pipeline[int] = 1
    pipeline[float] = 1.5
    with pytest.raises(sl.AmbiguousProvider):
        pipeline.compute(str)


def test_optional_dependency_can_be_filled_by_non_optional_param() -> None:
    def use_optional(x: Optional[int]) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_optional], params={int: 1})
    assert pipeline.compute(str) == '1'


def test_union_with_none_can_be_used_instead_of_Optional() -> None:
    def use_union1(x: Union[int, None]) -> str:
        return f'{x or 123}'

    def use_union2(x: Union[None, int]) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_union1], params={int: 1})
    assert pipeline.compute(str) == '1'
    pipeline = sl.Pipeline([use_union2], params={int: 1})
    assert pipeline.compute(str) == '1'


def test_optional_requested_directly_can_be_filled_by_non_optional_param() -> None:
    pipeline = sl.Pipeline([], params={int: 1})
    assert pipeline.compute(Optional[int]) == 1  # type: ignore[call-overload]


def test_optional_dependency_can_be_filled_transitively() -> None:
    def use_optional(x: Optional[int]) -> str:
        return f'{x or 123}'

    def make_int(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([use_optional, make_int], params={float: 2.2})
    assert pipeline.compute(str) == '2'


def test_optional_dependency_is_set_to_none_if_no_provider_found() -> None:
    def use_optional(x: Optional[int]) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_optional])
    assert pipeline.compute(str) == '123'


def test_optional_dependency_is_set_to_none_if_no_provider_found_transitively() -> None:
    def use_optional(x: Optional[int]) -> str:
        return f'{x or 123}'

    def make_int(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([use_optional, make_int])
    assert pipeline.compute(str) == '123'


def test_can_have_both_optional_and_non_optional_path_to_param() -> None:
    Str1 = NewType('Str1', str)
    Str2 = NewType('Str2', str)
    Str12 = NewType('Str12', str)
    Str21 = NewType('Str21', str)

    def use_optional_int(x: Optional[int]) -> Str1:
        return Str1(f'{x or 123}')

    def use_int(x: int) -> Str2:
        return Str2(f'{x}')

    def combine12(x: Str1, y: Str2) -> Str12:
        return Str12(f'{x} {y}')

    def combine21(x: Str2, y: Str1) -> Str21:
        return Str21(f'{x} {y}')

    pipeline = sl.Pipeline(
        [use_optional_int, use_int, combine12, combine21], params={int: 1}
    )
    assert pipeline.compute(Str12) == '1 1'
    assert pipeline.compute(Str21) == '1 1'


def test_presence_of_optional_does_not_affect_related_exception() -> None:
    Str1 = NewType('Str1', str)
    Str2 = NewType('Str2', str)
    Str12 = NewType('Str12', str)
    Str21 = NewType('Str21', str)

    def use_optional_int(x: Optional[int]) -> Str1:
        return Str1(f'{x or 123}')

    # Make sure the implementation does not unintentionally put "None" here,
    # triggered by the presence of the optional dependency on int in another provider.
    def use_int(x: int) -> Str2:
        return Str2(f'{x}')

    def combine12(x: Str1, y: Str2) -> Str12:
        return Str12(f'{x} {y}')

    def combine21(x: Str2, y: Str1) -> Str21:
        return Str21(f'{x} {y}')

    pipeline = sl.Pipeline([use_optional_int, use_int, combine12, combine21])
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(Str12)
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(Str21)


def test_optional_dependency_in_node_depending_on_param_table() -> None:
    def use_optional(x: float, y: Optional[int]) -> str:
        return f'{x} {y or 123}'

    pl = sl.Pipeline([use_optional])
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(sl.Series[int, str]) == sl.Series(
        int, {0: '1.0 123', 1: '2.0 123', 2: '3.0 123'}
    )
    pl[int] = 11
    assert pl.compute(sl.Series[int, str]) == sl.Series(
        int, {0: '1.0 11', 1: '2.0 11', 2: '3.0 11'}
    )


def test_optional_dependency_can_be_filled_from_param_table() -> None:
    def use_optional(x: Optional[float]) -> str:
        return f'{x or 4.0}'

    pl = sl.Pipeline([use_optional])
    pl.set_param_table(sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}))
    assert pl.compute(sl.Series[int, str]) == sl.Series(
        int, {0: '1.0', 1: '2.0', 2: '3.0'}
    )


def test_optional_without_anchoring_param_raises_when_requesting_series() -> None:
    Param = NewType('Param', float)

    def use_optional(x: Optional[float]) -> str:
        return f'{x or 4.0}'

    pl = sl.Pipeline([use_optional])
    pl.set_param_table(sl.ParamTable(int, {Param: [1.0, 2.0, 3.0]}))
    # It is a bit ambiguous what we would expect here: Above, we have another param
    # used from the table, defining the length of the series. Here, we could replicate
    # the output of use_optional(None) based on the `int` param table:
    #    sl.Series(int, {0: '4.0', 1: '4.0', 2: '4.0'})
    # However, we are not supporting this for non-optional dependencies either since
    # it is unclear whether that would bring conceptual issues or risk.
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(sl.Series[int, str])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import Any, Optional, Tuple, TypeVar, Union

from .scheduler import DaskScheduler, NaiveScheduler, Scheduler
from .typing import Graph, Item

T = TypeVar("T")


class TaskGraph:
    """
    Holds a concrete task graph and keys to compute.

    Task graphs are typically created by :py:class:`sciline.Pipeline.build`. They allow
    for computing all or a subset of the results in the graph.
    """

    def __init__(
        self,
        *,
        graph: Graph,
        keys: Union[type, Tuple[type, ...], Item[T], Tuple[Item[T], ...]],
        scheduler: Optional[Scheduler] = None,
    ) -> None:
        self._graph = graph
        self._keys = keys
        if scheduler is None:
            try:
                scheduler = DaskScheduler()
            except ImportError:
                scheduler = NaiveScheduler()
        self._scheduler = scheduler

    def compute(
        self,
        keys: Optional[
            Union[type, Tuple[type, ...], Item[T], Tuple[Item[T], ...]]
        ] = None,
    ) -> Any:
        """
        Compute the result of the graph.

        Parameters
        ----------
        keys:
            Optional list of keys to compute. This can be used to override the keys
            stored in the graph instance. Note that the keys must be present in the
            graph as intermediate results, otherwise KeyError is raised.

        Returns
        -------
        If ``keys`` is a single type, returns the single result that was computed.
        If ``keys`` is a tuple of types, returns a dictionary with type as keys
        and the corresponding results as values.

        """
        if keys is None:
            keys = self._keys
        if isinstance(keys, tuple):
            results = self._scheduler.get(self._graph, list(keys))
            return dict(zip(keys, results))
        else:
            return self._scheduler.get(self._graph, [keys])[0]

    def visualize(
        self, **kwargs: Any
    ) -> graphviz.Digraph:  # type: ignore[name-defined] # noqa: F821
        """
        Return a graphviz Digraph object representing the graph.

        Parameters
        ----------
        kwargs:
            Keyword arguments passed to :py:class:`graphviz.Digraph`.
        """
        from .visualize import to_graphviz

        return to_graphviz(self._graph, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from . import scheduler
from .domain import Scope, ScopeTwoParams
from .handler import (
    HandleAsBuildTimeException,
    HandleAsComputeTimeException,
    UnsatisfiedRequirement,
)
from .param_table import ParamTable
from .pipeline import AmbiguousProvider, Pipeline, UnboundTypeVar
from .series import Series

__all__ = [
    "AmbiguousProvider",
    "Series",
    "ParamTable",
    "Pipeline",
    "Scope",
    "ScopeTwoParams",
    "UnboundTypeVar",
    "UnsatisfiedRequirement",
    "scheduler",
    "HandleAsBuildTimeException",
    "HandleAsComputeTimeException",
]

del importlib



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType, TypeVar

import pytest

import sciline as sl

T = TypeVar("T")


def test_mypy_detects_wrong_arg_type_of_Scope_subclass() -> None:
    Param = TypeVar('Param')
    Param1 = NewType('Param1', int)

    class A(sl.Scope[Param, float], float):
        ...

    A[Param1](1.5)
    A[Param1]('abc')  # type: ignore[arg-type]


def test_missing_interface_of_scope_subclass_raises() -> None:
    param = TypeVar('param')

    with pytest.raises(TypeError, match="Missing or wrong interface for"):

        class A(sl.Scope[param, float]):
            ...


def test_mypy_accepts_interface_of_scope_sibling_class() -> None:
    param = TypeVar('param')
    param1 = NewType('param1', int)

    class A(sl.Scope[param, float], float):
        ...

    a = A[param1](1.5)
    a + a


def test_Scope_inconsistent_type_and_interface_raises() -> None:
    param = TypeVar('param')

    with pytest.raises(TypeError, match="Missing or wrong interface for"):

        class A(sl.Scope[param, str], float):
            ...


def test_ScopeTwoParams_inconsistent_type_and_interface_raises() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    with pytest.raises(TypeError, match="Missing or wrong interface for"):

        class A(sl.ScopeTwoParams[Param1, Param2, str], float):
            ...


def test_ScopeTwoParams() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    class A(sl.ScopeTwoParams[Param1, Param2, float], float):
        ...

    assert isinstance(A[int, int](1.5), float)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import Any, Callable, Generic, List, NewType, TypeVar

import numpy as np
import numpy.typing as npt
import pytest

import sciline as sl


def int_to_float(x: int) -> float:
    return 0.5 * x


def make_int() -> int:
    return 3


def int_float_to_str(x: int, y: float) -> str:
    return f"{x};{y}"


def test_pipeline_with_callables_can_compute_single_results() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    assert pipeline.compute(float) == 1.5
    assert pipeline.compute(int) == 3


def test_pipeline_does_not_autobind_types_that_can_be_default_constructed() -> None:
    # `int` can be constructed without arguments (and returns 0). Make sure that
    # the pipeline does not automatically bind `int` to `0`.
    pipeline = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(float)


def test_intermediate_used_multiple_times_is_computed_only_once() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    pipeline = sl.Pipeline([int_to_float, provide_int, int_float_to_str])
    assert pipeline.compute(str) == "3;1.5"
    assert ncall == 1


def test_multiple_keys_can_be_computed_without_repeated_calls() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    pipeline = sl.Pipeline([int_to_float, provide_int, int_float_to_str])
    assert pipeline.compute((float, str)) == {float: 1.5, str: "3;1.5"}
    assert ncall == 1


def test_multiple_keys_not_in_same_path_use_same_intermediate() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    def func1(x: int) -> float:
        return 0.5 * x

    def func2(x: int) -> str:
        return f"{x}"

    pipeline = sl.Pipeline([provide_int, func1, func2])
    assert pipeline.compute((float, str)) == {float: 1.5, str: "3"}
    assert ncall == 1


def test_Scope_subclass_can_be_set_as_param() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    pipeline = sl.Pipeline(params={Str[int]: Str[int]('1')})
    pipeline[Str[float]] = Str[float]('2.0')
    assert pipeline.compute(Str[int]) == Str[int]('1')
    assert pipeline.compute(Str[float]) == Str[float]('2.0')


def test_Scope_subclass_can_be_set_as_param_with_unbound_typevar() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    pipeline = sl.Pipeline()
    pipeline[Str[Param]] = Str[Param]('1')  # type: ignore[valid-type]
    assert pipeline.compute(Str[int]) == Str[int]('1')
    assert pipeline.compute(Str[float]) == Str[float]('1')


def test_ScopeTwoParam_subclass_can_be_set_as_param() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    class Str(sl.ScopeTwoParams[Param1, Param2, str], str):
        ...

    pipeline = sl.Pipeline(params={Str[int, float]: Str[int, float]('1')})
    pipeline[Str[float, int]] = Str[float, int]('2.0')
    assert pipeline.compute(Str[int, float]) == Str[int, float]('1')
    assert pipeline.compute(Str[float, int]) == Str[float, int]('2.0')


def test_ScopeTwoParam_subclass_can_be_set_as_param_with_unbound_typevar() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    class Str(sl.ScopeTwoParams[Param1, Param2, str], str):
        ...

    pipeline = sl.Pipeline()
    pipeline[Str[Param1, Param2]] = Str[Param1, Param2]('1')  # type: ignore[valid-type]
    assert pipeline.compute(Str[int, float]) == Str[int, float]('1')
    assert pipeline.compute(Str[float, int]) == Str[float, int]('1')


def test_generic_providers_produce_use_dependencies_based_on_bound_typevar() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized(x: Param) -> Str[Param]:
        return Str(f'{x}')

    def make_float() -> float:
        return 1.5

    def combine(x: Str[int], y: Str[float]) -> str:
        return f"{x};{y}"

    pipeline = sl.Pipeline([make_int, make_float, combine, parametrized])
    assert pipeline.compute(Str[int]) == Str[int]('3')
    assert pipeline.compute(Str[float]) == Str[float]('1.5')
    assert pipeline.compute(str) == '3;1.5'


def test_can_compute_result_depending_on_two_instances_of_generic_provider() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    Param = TypeVar('Param')

    class Float(sl.Scope[Param, float], float):
        ...

    class Str(sl.Scope[Param, str], str):
        ...

    def int_float_to_str(x: int, y: Float[Param]) -> Str[Param]:
        return Str(f"{x};{y}")

    Run1 = NewType('Run1', int)
    Run2 = NewType('Run2', int)
    Result = NewType('Result', str)

    def float1() -> Float[Run1]:
        return Float[Run1](1.5)

    def float2() -> Float[Run2]:
        return Float[Run2](2.5)

    def use_strings(s1: Str[Run1], s2: Str[Run2]) -> Result:
        return Result(f"{s1};{s2}")

    pipeline = sl.Pipeline(
        [provide_int, float1, float2, use_strings, int_float_to_str],
    )
    assert pipeline.compute(Result) == "3;1.5;3;2.5"
    assert ncall == 1


def test_subclasses_of_generic_provider_defined_with_Scope_work() -> None:
    Param = TypeVar('Param')

    class StrT(sl.Scope[Param, str], str):
        ...

    class Str1(StrT[Param]):
        ...

    class Str2(StrT[Param]):
        ...

    class Str3(StrT[Param]):
        ...

    class Str4(Str3[Param]):
        ...

    def make_str1() -> Str1[Param]:
        return Str1('1')

    def make_str2() -> Str2[Param]:
        return Str2('2')

    # Note that mypy cannot detect if when setting params, the type of the
    # parameter does not match the key. Same problem as with NewType.
    pipeline = sl.Pipeline(
        [make_str1, make_str2],
        params={
            Str3[int]: Str3[int]('int3'),
            Str3[float]: Str3[float]('float3'),
            Str4[int]: Str2[int]('int4'),
        },
    )
    assert pipeline.compute(Str1[float]) == Str1[float]('1')
    assert pipeline.compute(Str2[float]) == Str2[float]('2')
    assert pipeline.compute(Str3[int]) == Str3[int]('int3')
    assert pipeline.compute(Str3[float]) == Str3[float]('float3')
    assert pipeline.compute(Str4[int]) == Str4[int]('int4')


def test_subclasses_of_generic_array_provider_defined_with_Scope_work() -> None:
    Param = TypeVar('Param')

    class ArrayT(sl.Scope[Param, npt.NDArray[np.int64]], npt.NDArray[np.int64]):
        ...

    class Array1(ArrayT[Param]):
        ...

    class Array2(ArrayT[Param]):
        ...

    def make_array1() -> Array1[Param]:
        return Array1(np.array([1, 2, 3]))

    def make_array2() -> Array2[Param]:
        return Array2(np.array([4, 5, 6]))

    pipeline = sl.Pipeline([make_array1, make_array2])
    # Note that the param is not the dtype
    assert np.all(pipeline.compute(Array1[str]) == np.array([1, 2, 3]))
    assert np.all(pipeline.compute(Array2[str]) == np.array([4, 5, 6]))


def test_inserting_provider_returning_None_raises() -> None:
    def provide_none() -> None:
        return None

    with pytest.raises(ValueError):
        sl.Pipeline([provide_none])
    pipeline = sl.Pipeline([])
    with pytest.raises(ValueError):
        pipeline.insert(provide_none)


def test_inserting_provider_with_no_return_type_raises() -> None:
    def provide_none():  # type: ignore[no-untyped-def]
        return None

    with pytest.raises(ValueError):
        sl.Pipeline([provide_none])
    pipeline = sl.Pipeline([])
    with pytest.raises(ValueError):
        pipeline.insert(provide_none)


def test_TypeVar_requirement_of_provider_can_be_bound() -> None:
    T = TypeVar('T')

    def provider_int() -> int:
        return 3

    def provider(x: T) -> List[T]:
        return [x, x]

    pipeline = sl.Pipeline([provider_int, provider])
    assert pipeline.compute(List[int]) == [3, 3]


def test_TypeVar_that_cannot_be_bound_raises_UnboundTypeVar() -> None:
    T = TypeVar('T')

    def provider(_: T) -> int:
        return 1

    pipeline = sl.Pipeline([provider])
    with pytest.raises(sl.UnboundTypeVar):
        pipeline.compute(int)


def test_unsatisfiable_TypeVar_requirement_of_provider_raises() -> None:
    T = TypeVar('T')

    def provider_int() -> int:
        return 3

    def provider(x: T) -> List[T]:
        return [x, x]

    pipeline = sl.Pipeline([provider_int, provider])
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(List[float])


def test_TypeVar_params_are_not_associated_unless_they_match() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    class A(Generic[T1]):
        ...

    class B(Generic[T2]):
        ...

    def source() -> A[int]:
        return A[int]()

    def not_matching(x: A[T1]) -> B[T2]:
        return B[T2]()

    def matching(x: A[T1]) -> B[T1]:
        return B[T1]()

    pipeline = sl.Pipeline([source, not_matching])
    with pytest.raises(sl.UnboundTypeVar):
        pipeline.compute(B[int])

    pipeline = sl.Pipeline([source, matching])
    pipeline.compute(B[int])


def test_multi_Generic_with_fully_bound_arguments() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def source() -> A[int, float]:
        return A[int, float](1, 2.0)

    pipeline = sl.Pipeline([source])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)


def test_multi_Generic_with_partially_bound_arguments() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def source() -> float:
        return 2.0

    def partially_bound(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    pipeline = sl.Pipeline([source, partially_bound])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)


def test_multi_Generic_with_multiple_unbound() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def int_source() -> int:
        return 1

    def float_source() -> float:
        return 2.0

    def unbound(x: T1, y: T2) -> A[T1, T2]:
        return A[T1, T2](x, y)

    pipeline = sl.Pipeline([int_source, float_source, unbound])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)
    assert pipeline.compute(A[float, int]) == A[float, int](2.0, 1)


def test_distinct_fully_bound_instances_yield_distinct_results() -> None:
    T1 = TypeVar('T1')

    @dataclass
    class A(Generic[T1]):
        value: T1

    def int_source() -> A[int]:
        return A[int](1)

    def float_source() -> A[float]:
        return A[float](2.0)

    pipeline = sl.Pipeline([int_source, float_source])
    assert pipeline.compute(A[int]) == A[int](1)
    assert pipeline.compute(A[float]) == A[float](2.0)


def test_distinct_partially_bound_instances_yield_distinct_results() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def str_source() -> str:
        return 'a'

    def int_source(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    def float_source(x: T1) -> A[float, T1]:
        return A[float, T1](2.0, x)

    pipeline = sl.Pipeline([str_source, int_source, float_source])
    assert pipeline.compute(A[int, str]) == A[int, str](1, 'a')
    assert pipeline.compute(A[float, str]) == A[float, str](2.0, 'a')


def test_multiple_matching_partial_providers_raises() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def int_source() -> int:
        return 1

    def float_source() -> float:
        return 2.0

    def provider1(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    def provider2(x: T2) -> A[T2, float]:
        return A[T2, float](x, 2.0)

    pipeline = sl.Pipeline([int_source, float_source, provider1, provider2])
    assert pipeline.compute(A[int, int]) == A[int, int](1, 1)
    assert pipeline.compute(A[float, float]) == A[float, float](2.0, 2.0)
    with pytest.raises(sl.AmbiguousProvider):
        pipeline.compute(A[int, float])


def test_TypeVar_params_track_to_multiple_sources() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1]):
        value: T1

    @dataclass
    class B(Generic[T1]):
        value: T1

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2

    def provide_int() -> int:
        return 1

    def provide_float() -> float:
        return 2.0

    def provide_A(x: T1) -> A[T1]:
        return A[T1](x)

    # Note that it currently does not matter which TypeVar instance we use here:
    # Container tracks uses of TypeVar within a single provider, but does not carry
    # the information beyond the scope of a single call.
    def provide_B(x: T1) -> B[T1]:
        return B[T1](x)

    def provide_C(x: A[T1], y: B[T2]) -> C[T1, T2]:
        return C[T1, T2](x.value, y.value)

    pipeline = sl.Pipeline(
        [provide_int, provide_float, provide_A, provide_B, provide_C]
    )
    assert pipeline.compute(C[int, float]) == C[int, float](1, 2.0)


def test_instance_provider() -> None:
    Result = NewType('Result', float)

    def f(x: int, y: float) -> Result:
        return Result(x / y)

    pl = sl.Pipeline([f])
    pl[int] = 3
    pl[float] = 2.0
    assert pl.compute(int) == 3
    assert pl.compute(float) == 2.0
    assert pl.compute(Result) == 1.5


def test_provider_NewType_instance() -> None:
    A = NewType('A', int)
    pl = sl.Pipeline([])
    pl[A] = A(3)
    assert pl.compute(A) == 3


def test_setitem_generic_sets_up_working_subproviders() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A[int]] = A[int](3)
    pl[A[float]] = A[float](2.0)
    assert pl.compute(A[int]) == A[int](3)
    assert pl.compute(A[float]) == A[float](2.0)
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(A[str])


def test_setitem_generic_works_without_params() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A] = A(3)
    assert pl.compute(A) == A(3)


def test_setitem_raises_TypeError_if_instance_does_not_match_key() -> None:
    A = NewType('A', int)
    T = TypeVar('T')

    @dataclass
    class B(Generic[T]):
        value: T

    pl = sl.Pipeline()
    with pytest.raises(TypeError):
        pl[int] = 1.0
    with pytest.raises(TypeError):
        pl[A] = 1.0
    with pytest.raises(TypeError):
        pl[B[int]] = 1.0


def test_setitem_can_replace_param_with_param() -> None:
    pl = sl.Pipeline()
    pl[int] = 1
    pl[int] = 2
    assert pl.compute(int) == 2


def test_insert_can_replace_param_with_provider() -> None:
    def func() -> int:
        return 2

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func)
    assert pl.compute(int) == 2


def test_setitem_can_replace_provider_with_param() -> None:
    def func() -> int:
        return 2

    pl = sl.Pipeline()
    pl.insert(func)
    pl[int] = 1
    assert pl.compute(int) == 1


def test_insert_can_replace_provider_with_provider() -> None:
    def func1() -> int:
        return 1

    def func2() -> int:
        return 2

    pl = sl.Pipeline()
    pl.insert(func1)
    pl.insert(func2)
    assert pl.compute(int) == 2


def test_insert_can_replace_generic_provider_with_generic_provider() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func1(x: T) -> A[T]:
        return A[T](x)

    def func2(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func1)
    pl.insert(func2)
    assert pl.compute(A[int]) == A[int](2)


def test_insert_can_replace_generic_param_with_generic_provider() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)
    pl.insert(func)
    assert pl.compute(A[int]) == A[int](2)


def test_setitem_can_replace_generic_provider_with_generic_param() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func)
    assert pl.compute(A[int]) == A[int](2)
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)


def test_setitem_can_replace_generic_param_with_generic_param() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)
    pl[A[T]] = A[T](2)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](2)


def test_init_with_params() -> None:
    pl = sl.Pipeline(params={int: 1, float: 2.0})
    assert pl.compute(int) == 1
    assert pl.compute(float) == 2.0


def test_init_with_providers_and_params() -> None:
    def func(x: int, y: float) -> str:
        return f'{x} {y}'

    pl = sl.Pipeline(providers=[func], params={int: 1, float: 2.0})
    assert pl.compute(str) == "1 2.0"


def test_init_with_sciline_Scope_subclass_param_works() -> None:
    T = TypeVar('T')

    class A(sl.Scope[T, int], int):
        ...

    pl = sl.Pipeline(params={A[float]: A(1), A[str]: A(2)})
    assert pl.compute(A[float]) == A(1)
    assert pl.compute(A[str]) == A(2)


def test_building_graph_with_cycle_succeeds() -> None:
    def f(x: int) -> float:
        return float(x)

    def g(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([f, g])
    _ = pipeline.get(int)


def test_computing_graph_with_cycle_raises_CycleError() -> None:
    def f(x: int) -> float:
        return float(x)

    def g(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([f, g])
    with pytest.raises(sl.scheduler.CycleError):
        pipeline.compute(int)


def test_get_with_single_key_return_task_graph_that_computes_value() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int, int_float_to_str])
    task = pipeline.get(str)
    assert task.compute() == '3;1.5'


@pytest.mark.parametrize('key_type', [tuple, list, iter])
def test_get_with_key_iterable_return_task_graph_that_computes_dict_of_values(
    key_type: Callable[[Any], Any],
) -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(key_type((float, int)))
    assert task.compute() == {float: 1.5, int: 3}


def test_task_graph_compute_can_override_single_key() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float)
    assert task.compute(int) == 3


def test_task_graph_compute_can_override_key_tuple() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float)
    assert task.compute((int, float)) == {int: 3, float: 1.5}


def test_task_graph_compute_raises_if_override_keys_outside_graph() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(int)
    # The pipeline knows how to compute int, but the task graph does not
    # as the task graph is fixed at this point.
    with pytest.raises(KeyError):
        task.compute(float)


def test_get_with_NaiveScheduler() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float, scheduler=sl.scheduler.NaiveScheduler())
    assert task.compute() == 1.5


def test_bind_and_call_no_function() -> None:
    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(()) == ()


def test_bind_and_call_function_without_args() -> None:
    def func() -> str:
        return "func"

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == "func"


def test_bind_and_call_function_with_1_arg() -> None:
    def func(i: int) -> int:
        return i * 2

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == 6


def test_bind_and_call_function_with_2_arg2() -> None:
    def func(i: int, f: float) -> float:
        return i + f

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(func) == 4.5


def test_bind_and_call_overrides_default_args() -> None:
    def func(i: int, f: float = -0.5) -> float:
        return i + f

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(func) == 4.5


def test_bind_and_call_function_in_iterator() -> None:
    def func(i: int) -> int:
        return i * 2

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(iter((func,))) == (6,)


def test_bind_and_call_dataclass_without_args() -> None:
    @dataclass
    class C:
        ...

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(C) == C()


def test_bind_and_call_dataclass_with_1_arg() -> None:
    @dataclass
    class C:
        i: int

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(C) == C(i=3)


def test_bind_and_call_dataclass_with_2_arg2() -> None:
    @dataclass
    class C:
        i: int
        f: float

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(C) == C(i=3, f=1.5)


def test_bind_and_call_two_functions() -> None:
    def func1(i: int) -> int:
        return 2 * i

    def func2(f: float) -> float:
        return f + 1

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call((func1, func2)) == (6, 2.5)


def test_bind_and_call_two_functions_in_iterator() -> None:
    def func1(i: int) -> int:
        return 2 * i

    def func2(f: float) -> float:
        return f + 1

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(iter((func1, func2))) == (6, 2.5)


def test_bind_and_call_function_and_dataclass() -> None:
    def func(i: int) -> int:
        return 2 * i

    @dataclass
    class C:
        i: int
        f: float

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call((func, C)) == (6, C(i=3, f=1.5))


def test_bind_and_call_function_without_return_annotation() -> None:
    def func(i: int):  # type: ignore[no-untyped-def]
        return 2 * i

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == 6


def test_bind_and_call_generic_function() -> None:
    T = TypeVar('T')
    A = NewType('A', int)
    B = NewType('B', int)

    class G(sl.Scope[T, int], int):
        ...

    def func(a: G[A]) -> int:
        return -4 * a

    pipeline = sl.Pipeline([], params={G[A]: 3, G[B]: 4})
    assert pipeline.bind_and_call(func) == -12


def test_bind_and_call_function_runs_at_end() -> None:
    calls = []

    def a() -> int:
        calls.append('a')
        return 2

    def b() -> float:
        calls.append('b')
        return 3.1

    def c(_i: int) -> None:
        calls.append('c')

    def d(_f: float) -> None:
        calls.append('d')

    pipeline = sl.Pipeline([a, b])
    pipeline.bind_and_call([c, d])

    assert calls.index('a') in (0, 1)
    assert calls.index('b') in (0, 1)
    assert calls.index('c') in (2, 3)
    assert calls.index('d') in (2, 3)


def test_prioritizes_specialized_provider_over_generic() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    V = TypeVar('V', A, B)

    class H(sl.Scope[V, str], str):
        pass

    def p1(x: V) -> H[V]:
        return H[V]("Generic")

    def p2(x: B) -> H[B]:
        return H[B]("Special")

    pl = sl.Pipeline([p1, p2], params={A: 'A', B: 'B'})

    assert str(pl.compute(H[A])) == "Generic"
    assert str(pl.compute(H[B])) == "Special"


def test_prioritizes_specialized_provider_over_generic_several_typevars() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2
        third: str

    def p1(x: T1, y: T2) -> C[T1, T2]:
        return C(x, y, 'generic')

    def p2(x: A, y: T2) -> C[A, T2]:
        return C(x, y, 'medium generic')

    def p3(x: T2, y: B) -> C[T2, B]:
        return C(x, y, 'generic medium')

    def p4(x: A, y: B) -> C[A, B]:
        return C(x, y, 'special')

    pl = sl.Pipeline([p1, p2, p3, p4], params={A: A('A'), B: B('B')})

    assert pl.compute(C[B, A]) == C('B', 'A', 'generic')
    assert pl.compute(C[A, A]) == C('A', 'A', 'medium generic')
    assert pl.compute(C[B, B]) == C('B', 'B', 'generic medium')
    assert pl.compute(C[A, B]) == C('A', 'B', 'special')


def test_prioritizes_specialized_provider_raises() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2

    def p1(x: A, y: T1) -> C[A, T1]:
        return C(x, y)

    def p2(x: T1, y: B) -> C[T1, B]:
        return C(x, y)

    pl = sl.Pipeline([p1, p2], params={A: A('A'), B: B('B')})

    with pytest.raises(sl.AmbiguousProvider):
        pl.compute(C[A, B])

    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(C[B, A])


def test_compute_time_handler_allows_for_building_but_not_computing() -> None:
    def func(x: int) -> float:
        return 0.5 * x

    pipeline = sl.Pipeline([func])
    graph = pipeline.get(float, handler=sl.HandleAsComputeTimeException())
    with pytest.raises(sl.UnsatisfiedRequirement):
        graph.compute()


def test_pipeline_copy_simple() -> None:
    a = sl.Pipeline([int_to_float, make_int])
    b = a.copy()
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_dunder() -> None:
    a = sl.Pipeline([int_to_float, make_int])
    from copy import copy

    b = copy(a)
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_with_params() -> None:
    a = sl.Pipeline([int_to_float], params={int: 99})
    b = a.copy()
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5


def test_pipeline_copy_after_insert() -> None:
    a = sl.Pipeline([int_to_float])
    a.insert(make_int)
    b = a.copy()
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_after_setitem() -> None:
    a = sl.Pipeline([int_to_float])
    a[int] = 99
    b = a.copy()
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5


def test_copy_with_generic_providers() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized(x: Param) -> Str[Param]:
        return Str(f'{x}')

    def make_float() -> float:
        return 1.5

    def combine(x: Str[int], y: Str[float]) -> str:
        return f"{x};{y}"

    a = sl.Pipeline([make_int, make_float, combine, parametrized])
    b = a.copy()
    assert b.compute(Str[int]) == Str[int]('3')
    assert b.compute(Str[float]) == Str[float]('1.5')
    assert b.compute(str) == '3;1.5'


def test_pipeline_insert_on_copy_does_not_affect_original() -> None:
    a = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)
    b = a.copy()
    b.insert(make_int)
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)


def test_pipeline_insert_on_original_does_not_affect_copy() -> None:
    a = sl.Pipeline([int_to_float])
    b = a.copy()
    a.insert(make_int)
    assert a.compute(int) == 3
    assert a.compute(float) == 1.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        b.compute(int)


def test_pipeline_setitem_on_copy_does_not_affect_original() -> None:
    a = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)
    b = a.copy()
    b[int] = 99
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)


def test_pipeline_setitem_on_original_does_not_affect_copy() -> None:
    a = sl.Pipeline([int_to_float])
    b = a.copy()
    a[int] = 99
    assert a.compute(int) == 99
    assert a.compute(float) == 49.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        b.compute(int)


def test_pipeline_with_generics_setitem_on_original_does_not_affect_copy() -> None:
    RunType = TypeVar('RunType')

    class RawData(sl.Scope[RunType, int], int):
        ...

    class SquaredData(sl.Scope[RunType, int], int):
        ...

    Sample = NewType('Sample', int)
    Background = NewType('Background', int)
    Result = NewType('Result', int)

    def square(x: RawData[RunType]) -> SquaredData[RunType]:
        return SquaredData[RunType](x * x)

    def process(
        sample: SquaredData[Sample], background: SquaredData[Background]
    ) -> Result:
        return Result(sample + background)

    a = sl.Pipeline(
        [square, process],
        params={
            RawData[Sample]: 5,
            RawData[Background]: 2,
        },
    )
    assert a.compute(Result) == 29
    b = a.copy()
    assert b.compute(Result) == 29
    a[RawData[Sample]] = 7
    assert a.compute(Result) == 53
    assert b.compute(Result) == 29


def test_pipeline_with_generics_setitem_on_copy_does_not_affect_original() -> None:
    RunType = TypeVar('RunType')

    class RawData(sl.Scope[RunType, int], int):
        ...

    class SquaredData(sl.Scope[RunType, int], int):
        ...

    Sample = NewType('Sample', int)
    Background = NewType('Background', int)
    Result = NewType('Result', int)

    def square(x: RawData[RunType]) -> SquaredData[RunType]:
        return SquaredData[RunType](x * x)

    def process(
        sample: SquaredData[Sample], background: SquaredData[Background]
    ) -> Result:
        return Result(sample + background)

    a = sl.Pipeline(
        [square, process],
        params={
            RawData[Sample]: 5,
            RawData[Background]: 2,
        },
    )
    assert a.compute(Result) == 29
    b = a.copy()
    assert b.compute(Result) == 29
    b[RawData[Sample]] = 7
    assert a.compute(Result) == 29
    assert b.compute(Result) == 53


def test_html_repr() -> None:
    pipeline = sl.Pipeline([make_int], params={float: 5.0})
    assert isinstance(pipeline._repr_html_(), str)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import Any, Collection, Dict, Mapping, Optional


class ParamTable(Mapping[type, Collection[Any]]):
    """A table of parameters with a row index and named row dimension."""

    def __init__(
        self,
        row_dim: type,
        columns: Dict[type, Collection[Any]],
        *,
        index: Optional[Collection[Any]] = None,
    ):
        """
        Create a new param table.

        Parameters
        ----------
        row_dim:
            The row dimension. This must be a type or a type-alias (not an instance),
            and is used by :py:class:`sciline.Pipeline` to identify each parameter
            table.
        columns:
            The columns of the table. The keys (column names) must be types or type-
            aliases matching the values in the respective columns.
        index:
            The row index of the table. If not given, a default index will be
            generated, as the integer range of the column length.
        """
        sizes = set(len(v) for v in columns.values())
        if len(sizes) != 1:
            raise ValueError(
                f"Columns in param table must all have same size, got {sizes}"
            )
        size = sizes.pop()
        if index is not None:
            if len(index) != size:
                raise ValueError(
                    f"Index length not matching columns, got {len(index)} and {size}"
                )
            if len(set(index)) != len(index):
                raise ValueError(f"Index must be unique, got {index}")
        self._row_dim = row_dim
        self._columns = columns
        self._index = index or list(range(size))

    @property
    def row_dim(self) -> type:
        """The row dimension of the table."""
        return self._row_dim

    @property
    def index(self) -> Collection[Any]:
        """The row index of the table."""
        return self._index

    def __contains__(self, key: Any) -> bool:
        return self._columns.__contains__(key)

    def __getitem__(self, key: Any) -> Any:
        return self._columns.__getitem__(key)

    def __iter__(self) -> Any:
        return self._columns.__iter__()

    def __len__(self) -> int:
        return self._columns.__len__()

    def __repr__(self) -> str:
        return f"ParamTable(row_dim={self.row_dim.__name__}, columns={self._columns})"

    def _repr_html_(self) -> str:
        return (
            f"<table><tr><th>{self.row_dim.__name__}</th>"
            + "".join(
                f"<th>{getattr(k, '__name__', str(k).split('.')[-1])}</th>"
                for k in self._columns.keys()
            )
            + "</tr>"
            + "".join(
                f"<tr><td>{idx}</td>" + "".join(f"<td>{v}</td>" for v in row) + "</tr>"
                for idx, row in zip(self.index, zip(*self._columns.values()))
            )
            + "</table>"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest

import sciline as sl
from sciline.task_graph import TaskGraph
from sciline.typing import Graph


def as_float(x: int) -> float:
    return 0.5 * x


def make_task_graph() -> Graph:
    pl = sl.Pipeline([as_float], params={int: 1})
    return pl.build(float, handler=sl.HandleAsBuildTimeException())


def test_default_scheduler_is_dask_when_dask_available() -> None:
    _ = pytest.importorskip("dask")
    tg = TaskGraph(graph={}, keys=())
    assert isinstance(tg._scheduler, sl.scheduler.DaskScheduler)


def test_compute_returns_value_when_initialized_with_single_key() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, keys=float)
    assert tg.compute() == 0.5


def test_compute_returns_dict_when_initialized_with_key_tuple() -> None:
    graph = make_task_graph()
    assert TaskGraph(graph=graph, keys=(float,)).compute() == {float: 0.5}
    assert TaskGraph(graph=graph, keys=(float, int)).compute() == {float: 0.5, int: 1}


def test_compute_returns_value_when_provided_with_single_key() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, keys=float)
    assert tg.compute(int) == 1


def test_compute_returns_dict_when_provided_with_key_tuple() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, keys=float)
    assert tg.compute((int, float)) == {int: 1, float: 0.5}


def test_compute_raises_when_provided_with_key_not_in_graph() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, keys=float)
    with pytest.raises(KeyError):
        tg.compute(str)
    with pytest.raises(KeyError):
        tg.compute((str, float))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest

import sciline as sl


def test_raises_with_zero_columns() -> None:
    with pytest.raises(ValueError):
        sl.ParamTable(row_dim=int, columns={})


def test_raises_with_inconsistent_column_sizes() -> None:
    with pytest.raises(ValueError):
        sl.ParamTable(row_dim=int, columns={int: [1, 2, 3], float: [1.0, 2.0]})


def test_raises_with_inconsistent_index_length() -> None:
    with pytest.raises(ValueError):
        sl.ParamTable(row_dim=int, columns={float: [1.0, 2.0]}, index=[1, 2, 3])


def test_raises_with_non_unique_index() -> None:
    with pytest.raises(ValueError):
        sl.ParamTable(int, {float: [1.0, 2.0, 3.0]}, index=[1, 1, 2])


def test_contains_includes_all_columns() -> None:
    pt = sl.ParamTable(row_dim=int, columns={int: [1, 2, 3], float: [1.0, 2.0, 3.0]})
    assert int in pt
    assert float in pt
    assert str not in pt


def test_contains_does_not_include_index() -> None:
    pt = sl.ParamTable(row_dim=int, columns={float: [1.0, 2.0, 3.0]})
    assert int not in pt


def test_len_is_number_of_columns() -> None:
    pt = sl.ParamTable(row_dim=int, columns={int: [1, 2, 3], float: [1.0, 2.0, 3.0]})
    assert len(pt) == 2


def test_defaults_to_range_index() -> None:
    pt = sl.ParamTable(row_dim=int, columns={float: [1.0, 2.0, 3.0]})
    assert pt.index == [0, 1, 2]



from collections import defaultdict
from typing import Any, Callable, DefaultDict, Iterable, TypeVar

T = TypeVar('T')
G = TypeVar('G')


def groupby(f: Callable[[T], G], a: Iterable[T]) -> DefaultDict[G, list[T]]:
    g = defaultdict(lambda: [])
    for e in a:
        g[f(e)].append(e)
    return g


def qualname(obj: Any) -> str:
    return str(
        obj.__qualname__ if hasattr(obj, '__qualname__') else obj.__class__.__qualname__
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Any, Generic, TypeVar, get_args, get_origin

PARAM = TypeVar("PARAM")
PARAM2 = TypeVar("PARAM2")
SUPER = TypeVar("SUPER")


def _check_supertype(cls: type, scope_cls: type) -> None:
    # Mypy does not support __orig_bases__ yet(?)
    # See also https://stackoverflow.com/a/73746554 for useful info
    scope = cls.__orig_bases__[0]  # type: ignore[attr-defined]
    # Only check direct subclasses
    if get_origin(scope) is scope_cls:
        supertype = get_args(scope)[-1]
        # Remove potential generic params
        # In Python 3.8, get_origin does not work with numpy.typing.NDArray,
        # but it defines __origin__
        supertype = getattr(supertype, '__origin__', None) or supertype
        if supertype not in cls.__bases__:
            raise TypeError(
                f"Missing or wrong interface for {cls}, "
                f"should inherit {supertype}.\n"
                "Example:\n"
                "\n"
                "    Param = TypeVar('Param')\n"
                "    \n"
                "    class A(sl.Scope[Param, float], float):\n"
                "        ...\n"
            )


class Scope(Generic[PARAM, SUPER]):
    """
    Helper for defining a generic type alias.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _check_supertype(cls, Scope)
        return super().__init_subclass__(**kwargs)

    def __new__(cls, x: SUPER) -> SUPER:  # type: ignore[misc]
        return x


# Can hopefully remove this once we can use TypeVarTuple
class ScopeTwoParams(Generic[PARAM, PARAM2, SUPER]):
    """
    Helper for defining a generic type alias.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _check_supertype(cls, ScopeTwoParams)
        return super().__init_subclass__(**kwargs)

    def __new__(cls, x: SUPER) -> SUPER:  # type: ignore[misc]
        return x



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest

import sciline as sl

dask = pytest.importorskip("dask")


def test_dask_scheduler_is_threaded_by_default() -> None:
    scheduler = sl.scheduler.DaskScheduler()
    assert scheduler._dask_get == dask.threaded.get



import inspect
from dataclasses import dataclass
from html import escape
from itertools import chain
from typing import Any, Literal, Mapping, Optional, Sequence, Tuple, TypeVar, Union

from .typing import Item, Key
from .utils import groupby, qualname

ProviderKind = Literal['function', 'parameter', 'table']


@dataclass
class ProviderDisplayData:
    origin: Key
    args: Tuple[Union[Key, TypeVar], ...]
    kind: ProviderKind
    value: Any


def _details(summary: str, body: str) -> str:
    return f'''
    <details>
      <summary>{summary}</summary>
      {body}
    </details>
    '''


def _provider_name(p: Any) -> str:
    if isinstance(p, tuple):
        (name, cname), values = p
        return escape(f'{qualname(cname)}({qualname(name)})')
    name = f'{qualname(p.origin)}'
    if p.args:
        args = ','.join(
            ('*' if isinstance(arg, TypeVar) else f'{qualname(arg)}' for arg in p.args)
        )
        name += f'[{args}]'
    return escape(f'{name}')


def _provider_source(p: Any) -> str:
    if isinstance(p, tuple):
        (name, cname), values = p
        return escape(f'ParamTable({qualname(name)}, length={len(values)})')
    if p.kind == 'function':
        module = getattr(inspect.getmodule(p.value), '__name__', '')
        return _details(
            escape(p.value.__name__),
            escape(f'{module}.{p.value.__name__}'),
        )
    return ''


def _provider_value(p: Any) -> str:
    if not isinstance(p, tuple) and p.kind == 'parameter':
        html = escape(str(p.value)).strip()
        return _details(f'{html[:30]}...', html) if len(html) > 30 else html
    return ''


def pipeline_html_repr(
    providers: Mapping[ProviderKind, Sequence[ProviderDisplayData]]
) -> str:
    def table_name_and_column_name(p: ProviderDisplayData) -> Optional[Tuple[Any, Any]]:
        if isinstance(p.origin, Item):
            return (p.origin.label[0].tp, p.origin.tp)
        return None

    param_table_columns_by_name_colname = groupby(
        table_name_and_column_name,
        providers['table'],
    )
    provider_rows = '\n'.join(
        (
            f'''
        <tr>
          <td scope="row">{_provider_name(p)}</td>
          <td scope="row">{_provider_value(p)}</td>
          <td scope="row">{_provider_source(p)}</th>
        </tr>'''
            for p in sorted(
                chain(
                    providers['function'],
                    providers['parameter'],
                    param_table_columns_by_name_colname.items(),
                ),
                key=_provider_name,
            )
        )
    )
    return f'''
    <div class="pipeline-html-repr">
      <table>
        <thead>
          <tr>
            <th scope="col">Name</th>
            <th scope="col">Value</th>
            <th scope="col">Source</th>
          </tr>
        </thead>
        <tbody>
          {provider_rows}
        </tbody>
      </table>
    </div>
    '''.strip()
